#include <iostream>

struct node;

struct node_ptr {
	node* ptr;
	node_ptr(node* ptr = nullptr) : ptr(ptr) {}

	node& operator*() const { return *ptr; }

	// no copies!
	node_ptr(const node_ptr&) = delete;
	node_ptr& operator=(const node_ptr&) = delete;

	// moves are OK
	node_ptr(node_ptr&& other) noexcept
	  : ptr(other.ptr) {
		other.ptr = nullptr;
	}

	node_ptr& operator=(node_ptr&& other) noexcept {
		release();
		ptr = other.ptr;
		other.ptr = nullptr;
		return *this;
	}

	// using delete before the actual definition of node might be UB
	~node_ptr() { release(); }

	void release();
};

struct node {
	int value;
	node_ptr next;
};

void node_ptr::release() {
	// cleanup
	delete ptr;
}

struct list_iter {
	node* current;
	int& operator*() const { return current->value; }
	list_iter& operator++() {
		current = current->next.ptr;
		return *this;
	}
	// use pointer comparison
	friend bool operator==(list_iter, list_iter) = default;
};

struct list {
	node_ptr first;
	list_iter begin() { return { first.ptr }; }
	list_iter end() { return { nullptr }; }

	void insert_at_start(int value) {
		first = { new node{ value, std::move(first) } };
	}
};

int main() {
	list lst;
	for (int i = 1; i < 10; ++i) {
		lst.insert_at_start(i);
	}

	for (int x : lst) {
		std::cout << x << '\n';
	}
}
