#include <iostream>
#include <memory>

struct thing {
	int id;

	thing(int id) : id(id) {
		std::cout << id << " created\n";
	}
	~thing() {
		std::cout << id << " destroyed\n";
	}

	// no copies, no moves
	thing(const thing&) = delete;
	thing(thing&&) = delete;
	thing& operator=(const thing&) = delete;
	thing& operator=(thing&&) = delete;

	void do_something() const {
		std::cout << id << " doing something...\n";
	}
};

void run(const thing& thing) {
	thing.do_something();
}

void run_with_ptr(const thing* ptr) {
	if (ptr) {
		ptr->do_something();
	} else {
		std::cout << "pointer is null\n";
	}
}

int main() {
	std::cout << "before inner block\n";
	{
		// C++14
		auto ptr = std::make_shared<thing>(17);
		std::cout << "inside inner block\n";

		ptr->do_something();
		run(*ptr);

		std::cout << "still inside inner block\n";
	}
	std::cout << "after inner block\n";

	auto ptr1 = std::make_shared<thing>(1);
	auto ptr2 = ptr1;
	std::cout << "two shared owners, reset one\n";
	ptr1.reset();
	std::cout << "reset the other\n";
	ptr2.reset();
	std::cout << "---\n";

	{
		auto ptr = std::make_shared<thing>(7);
		ptr1 = ptr;
		ptr2 = ptr;
		std::cout << "use_count: " << ptr.use_count() << '\n';
	}
	std::cout << "---\n";
}
