#include <iostream>
#include <memory>

struct thing {
	int id;

	thing(int id) : id(id) {
		std::cout << id << " created\n";
	}
	~thing() {
		std::cout << id << " destroyed\n";
	}

	// no copies, no moves
	thing(const thing&) = delete;
	thing(thing&&) = delete;
	thing& operator=(const thing&) = delete;
	thing& operator=(thing&&) = delete;

	void do_something() const {
		std::cout << id << " doing something...\n";
	}
};

void run(const thing& thing) {
	thing.do_something();
}

void run_with_ptr(const thing* ptr) {
	if (ptr) {
		ptr->do_something();
	} else {
		std::cout << "pointer is null\n";
	}
}

int main() {
	std::cout << "before inner block\n";
	{
		// pre-C++14; prefer the new version
		// std::unique_ptr<thing> ptr(new thing(7));

		// C++14
		auto ptr = std::make_unique<thing>(17);
		std::cout << "inside inner block\n";

		ptr->do_something();
		run(*ptr);

		std::cout << "still inside inner block\n";
	}
	std::cout << "after inner block\n";

	auto ptr1 = std::make_unique<thing>(1);
	auto ptr2 = std::make_unique<thing>(2);

	std::cout << "before move\n";
	ptr1 = std::move(ptr2);
	ptr1->do_something();
	if (ptr2 == nullptr) {
		std::cout << "ptr2 is null\n";
	}
	std::cout << "after move\n";
	ptr1.reset();
	std::cout << "after reset\n";

	auto ptr3 = std::make_unique<thing>(3);
	run_with_ptr(ptr3.get());
	ptr3 = nullptr;
	run_with_ptr(ptr3.get());

	std::cout << "end of main\n";
}
